import { Socket as ClientSocket, io } from 'socket.io-client'
import { Server, Socket as SocketForConnectedClient } from "socket.io"
import { Subject } from "rxjs";
import { createServer } from "http";

let fromServer = new Subject<{ event: 'profile' | 'message', payload: any }>()
let toServer = new Subject<{ event: 'profile' | 'message', payload: any }>()

startSocketServer(3001)
// startSocketServer(3002)
startClientSocketConnection('http://localhost:3000')
consoleLog()

function consoleLog(): void {
    fromServer.subscribe(message => console.log(`From Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
    toServer.subscribe(message => console.log(`To Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
}

function startSocketServer(port: number): void {
    console.log(`Socket Server ${port} Started....`)
    let httpServer = createServer();
    let socketServer = new Server(httpServer)

    socketServer.on('connection', (socket: SocketForConnectedClient) => {
        socket.on(`profile`, (msg) => {
            toServer.next({ event: 'profile', payload: msg })
        })

        socket.on('message', (msg) => {
            toServer.next({ event: 'message', payload: msg })
        })

        socket.on(`disconnect`, () => {
            console.log(`I shouldn't be here lol`)
        })

        fromServer.subscribe(message => {
            socketServer.emit(message.event, message.payload)
        })
    })

    socketServer.engine.on("connection_error", (err) => {
        console.log(err.req);      // the request object
        console.log(err.code);     // the error code, for example 1
        console.log(err.message);  // the error message, for example "Session ID unknown"
        console.log(err.context);  // some additional error context
    });

    // Start the HTTP server on 127.0.0.1 with the given port
    httpServer.listen(port, '0.0.0.0', () => {
        console.log(`Socket server listening on ${port}`);
    });

}

function startClientSocketConnection(serverUrl: string): void {
    // let clientSocket = io(serverUrl)
    let clientSocket: ClientSocket = io(serverUrl, {
        reconnection: true,              // Enable automatic reconnections
        reconnectionAttempts: 1000,       // Retry up to 10 times
        reconnectionDelay: 500,          // Start with a 500ms delay
        reconnectionDelayMax: 10000,     // Delay can grow to a max of 10 seconds
        randomizationFactor: 0.3,
    })

    toServer.subscribe(message => { clientSocket.emit(message.event, message.payload) })

    clientSocket.on(`connect`, () => {

    })

    clientSocket.on(`profile`, (msg) => {
        fromServer.next({ event: 'profile', payload: msg })
    })

    clientSocket.on(`message`, (msg) => {
        fromServer.next({ event: 'message', payload: msg })
    })

    clientSocket.on(`disconnect`, () => {

    })
}

